/* Sample code using the Chunks and Tasks programming model.
 * See <http://www.chunks-and-tasks.org>.
 */
#include <stdexcept>
#include "SmithWaterman.h"


static int max4(int a1, int a2, int a3, int a4) {
  if(a1 > a2 && a1 > a3 && a1 > a4)
    return a1;
  if(a2 > a3 && a2 > a4)
    return a2;
  if(a3 > a4)
    return a3;
  return a4;
}

struct GetRight: public cht::Task {
  cht::ID execute(SWResult const & x);
  CHT_TASK_INPUT((SWResult));
  CHT_TASK_OUTPUT((chttl::ChunkVector<int>));
  CHT_TASK_TYPE_DECLARATION;
};
CHT_TASK_TYPE_IMPLEMENTATION((GetRight));
cht::ID GetRight::execute(SWResult const & x) {
  return x.get_cid_H_rightmost_column();
}

struct GetBottom: public cht::Task {
  cht::ID execute(SWResult const & x);
  CHT_TASK_INPUT((SWResult));
  CHT_TASK_OUTPUT((chttl::ChunkVector<int>));
  CHT_TASK_TYPE_DECLARATION;
};
CHT_TASK_TYPE_IMPLEMENTATION((GetBottom));
cht::ID GetBottom::execute(SWResult const & x) {
  return x.get_cid_H_bottom_row();
}

struct GetLastElement: public cht::Task {
  cht::ID execute(chttl::ChunkVector<int> const & x);
  CHT_TASK_INPUT((chttl::ChunkVector<int>));
  CHT_TASK_OUTPUT((chttl::Basic<int>));
  CHT_TASK_TYPE_DECLARATION;
};
CHT_TASK_TYPE_IMPLEMENTATION((GetLastElement));
cht::ID GetLastElement::execute(chttl::ChunkVector<int> const & x) {
  int N = x.size();
  assert(N > 0);
  int value = x[N-1];
  return registerChunk<chttl::Basic<int> >(new chttl::Basic<int>(value));
}

struct CombineResultsFinal: public cht::Task {
  cht::ID execute(chttl::Basic<int> const & maxValue, cht::ChunkID const & cid_H_bottom_row, cht::ChunkID const & cid_H_rightmost_column);
  CHT_TASK_INPUT((chttl::Basic<int>, cht::ChunkID, cht::ChunkID));
  CHT_TASK_OUTPUT((SWResult));
  CHT_TASK_TYPE_DECLARATION;
};
CHT_TASK_TYPE_IMPLEMENTATION((CombineResultsFinal));
cht::ID CombineResultsFinal::execute(chttl::Basic<int> const & maxValue, cht::ChunkID const & cid_H_bottom_row, cht::ChunkID const & cid_H_rightmost_column) {
  SWResult* result = new SWResult();
  result->maxValue = maxValue;
  result->setChildren(cid_H_bottom_row, cid_H_rightmost_column);
  return registerChunk<SWResult>(result);
}

struct CombineResults: public cht::Task {
  cht::ID execute(SWResult const & x11, SWResult const & x12, SWResult const & x21, SWResult const & x22);
  CHT_TASK_INPUT((SWResult, SWResult, SWResult, SWResult));
  CHT_TASK_OUTPUT((SWResult));
  CHT_TASK_TYPE_DECLARATION;
};
CHT_TASK_TYPE_IMPLEMENTATION((CombineResults));
cht::ID CombineResults::execute(SWResult const & x11, SWResult const & x12, SWResult const & x21, SWResult const & x22) {
  int maxValue = max4(x11.maxValue, x12.maxValue, x21.maxValue, x22.maxValue);
  cht::ChunkID cid_maxValue = registerChunk<chttl::Basic<int> >(new chttl::Basic<int>(maxValue));
  cht::ID id_H_bottom_row = registerTask<chttl::ChunkVectorConcatenate<int> >(x21.get_cid_H_bottom_row(), x22.get_cid_H_bottom_row());
  cht::ID id_H_rightmost_column = registerTask<chttl::ChunkVectorConcatenate<int> >(x12.get_cid_H_rightmost_column(), x22.get_cid_H_rightmost_column());
  return registerTask<CombineResultsFinal>(cid_maxValue, id_H_bottom_row, id_H_rightmost_column);
}


CHT_TASK_TYPE_IMPLEMENTATION((SW));

static int do_SW_explicitly(chttl::ChunkVector<char> const & seqA,
			    chttl::ChunkVector<char> const & seqB,
			    chttl::ChunkVector<int> const & leftBorder,
			    chttl::ChunkVector<int> const & topBorder,
			    chttl::Basic<int> const & topLeftValue,
			    std::vector<int> & H_bottom_row,
			    std::vector<int> & H_rightmost_column) {
  int N = seqA.size();
  const int similarity_score_for_match = 2;
  const int similarity_score_for_mismatch = -1;
  const int gap_penalty_d = 2;
  std::vector<int> H(N*N);
  int maxValue = 0;
  for(int startRow = 0; startRow < 2*N-1; startRow++) {
    for(int col = 0; col <= startRow; col++) {
      int row = startRow - col;
      assert(row >= 0);
      if(row >= N)
	continue;
      if(col >= N)
	continue;
      int H_topleft = topLeftValue;
      if(row > 0 && col > 0)
	H_topleft = H[(row-1)*N+(col-1)];
      if(row > 0 && col == 0)
	H_topleft = leftBorder[row-1];
      if(row == 0 && col > 0)
	H_topleft = topBorder[col-1];
      int H_top = topBorder[col];
      if(row > 0)
	H_top = H[(row-1)*N+(col-0)];
      int H_left = leftBorder[row];
      if(col > 0)
	H_left = H[(row-0)*N+(col-1)];
      int S_similarity_score = similarity_score_for_mismatch;
      if(seqA[row] == seqB[col])
	S_similarity_score = similarity_score_for_match;
      // Take max of 4 alternatives
      int alt1 = H_topleft + S_similarity_score;
      int alt2 = H_left - gap_penalty_d;
      int alt3 = H_top - gap_penalty_d;
      int alt4 = 0;
      int Hvalue = max4(alt1, alt2, alt3, alt4);
      H[row*N+col] = Hvalue;
      if(Hvalue > maxValue)
	maxValue = Hvalue;
    }
  }
  H_bottom_row.resize(N);
  H_rightmost_column.resize(N);
  for(int i = 0; i < N; i++) {
    H_bottom_row[i] = H[(N-1)*N+i];
    H_rightmost_column[i] = H[i*N+(N-1)];
  }
  return maxValue;
}

template <typename T>
void split_vector(const std::vector<T> & v, std::vector<T> & v_part1, std::vector<T> & v_part2) {
  int N = v.size();
  v_part1.resize(N/2);
  v_part2.resize(N/2);
  for(int i = 0; i < N/2; i++)
    v_part1[i] = v[i];
  for(int i = 0; i < N/2; i++)
    v_part2[i] = v[N/2+i];
}

cht::ID SW::execute(chttl::ChunkVector<char> const & seqA,
		    chttl::ChunkVector<char> const & seqB,
		    chttl::ChunkVector<int> const & leftBorder,
		    chttl::ChunkVector<int> const & topBorder,
		    chttl::Basic<int> const & topLeftValue) {
  assert(seqA.size() == seqB.size());
  assert(leftBorder.size() == topBorder.size());
  assert(seqA.size() == leftBorder.size());
  int N = seqA.size();
  // Check if we are at lowest level
  if(N < 4) {
    // Lowest level
    std::vector<int> H_bottom_row;
    std::vector<int> H_rightmost_column;
    int maxValue = do_SW_explicitly(seqA,
				    seqB,
				    leftBorder,
				    topBorder,
				    topLeftValue,
				    H_bottom_row,
				    H_rightmost_column);
    // Prepare result chunk object
    SWResult* result = new SWResult();
    result->maxValue = maxValue;
    cht::ChunkID cid_H_bottom_row       = registerChunk<chttl::ChunkVector<int> >(new chttl::ChunkVector<int>(H_bottom_row));
    cht::ChunkID cid_H_rightmost_column = registerChunk<chttl::ChunkVector<int> >(new chttl::ChunkVector<int>(H_rightmost_column));
    result->setChildren(cid_H_bottom_row, cid_H_rightmost_column);
    return registerChunk<SWResult>(result);
  }
  else {
    // Not lowest level
    assert(N % 2 == 0);
    // Divide sequences into 2 parts
    std::vector<char> seqA_part1, seqA_part2;
    split_vector<char>(seqA, seqA_part1, seqA_part2);
    cht::ChunkID cid_seq_A_part1 = registerChunk<chttl::ChunkVector<char> >(new chttl::ChunkVector<char>(seqA_part1));
    cht::ChunkID cid_seq_A_part2 = registerChunk<chttl::ChunkVector<char> >(new chttl::ChunkVector<char>(seqA_part2));
    std::vector<char> seqB_part1, seqB_part2;
    split_vector<char>(seqB, seqB_part1, seqB_part2);
    cht::ChunkID cid_seq_B_part1 = registerChunk<chttl::ChunkVector<char> >(new chttl::ChunkVector<char>(seqB_part1));
    cht::ChunkID cid_seq_B_part2 = registerChunk<chttl::ChunkVector<char> >(new chttl::ChunkVector<char>(seqB_part2));
    // Divide borders into 2 parts
    std::vector<int> leftBorder_part1, leftBorder_part2;
    split_vector<int>(leftBorder, leftBorder_part1, leftBorder_part2);
    cht::ChunkID cid_leftBorder_part1 = registerChunk<chttl::ChunkVector<int> >(new chttl::ChunkVector<int>(leftBorder_part1));
    cht::ChunkID cid_leftBorder_part2 = registerChunk<chttl::ChunkVector<int> >(new chttl::ChunkVector<int>(leftBorder_part2));
    std::vector<int> topBorder_part1, topBorder_part2;
    split_vector<int>(topBorder, topBorder_part1, topBorder_part2);
    cht::ChunkID cid_topBorder_part1 = registerChunk<chttl::ChunkVector<int> >(new chttl::ChunkVector<int>(topBorder_part1));
    cht::ChunkID cid_topBorder_part2 = registerChunk<chttl::ChunkVector<int> >(new chttl::ChunkVector<int>(topBorder_part2));
    // Divide work into 4 parts, like this:
    // 11 12
    // 21 22
    // First do part 11
    cht::ID id_11 = registerTask<SW>(cid_seq_A_part1, cid_seq_B_part1, cid_leftBorder_part1, cid_topBorder_part1, getInputChunkID(topLeftValue));
    // Get borders from 11 result, to use as input for parts 12 and 21
    cht::ID id_11_right = registerTask<GetRight>(id_11);
    cht::ID id_11_bottom = registerTask<GetBottom>(id_11);
    // Now do part 12
    cht::ChunkID cid_top_left_value_for_12 = registerChunk<chttl::Basic<int> >(new chttl::Basic<int>(topBorder[N/2-1]));
    cht::ID id_12 = registerTask<SW>(cid_seq_A_part1, cid_seq_B_part2, id_11_right, cid_topBorder_part2, cid_top_left_value_for_12);
    // Now do part 21
    int top_left_value_for_21 = leftBorder[N/2-1];
    cht::ChunkID cid_top_left_value_for_21 = registerChunk<chttl::Basic<int> >(new chttl::Basic<int>(top_left_value_for_21));
    cht::ID id_21 = registerTask<SW>(cid_seq_A_part2, cid_seq_B_part1, cid_leftBorder_part2, id_11_bottom, cid_top_left_value_for_21);
    // Get borders from 12 and 21 results, to use as input for part 22
    cht::ID id_21_right = registerTask<GetRight>(id_21);
    cht::ID id_12_bottom = registerTask<GetBottom>(id_12);
    // Now do part 22
    cht::ID id_top_left_value_for_22 = registerTask<GetLastElement>(id_11_bottom);
    cht::ID id_22 = registerTask<SW>(cid_seq_A_part2, cid_seq_B_part2, id_21_right, id_12_bottom, id_top_left_value_for_22);
    // Now we have done all 4 parts; combine results
    return registerTask<CombineResults>(id_11, id_12, id_21, id_22);
  }
}
