/* Sample code using the Chunks and Tasks programming model.
 * See <http://www.chunks-and-tasks.org>.
 */
#include <iostream>
#include <cstdio>
#include <cstring>
#include <cassert>
#include <vector>

static void read_sequence_from_file(const char* fileName, std::vector<char> & resultSeq) {
  FILE* inFile = fopen(fileName, "rb");
  // Get filesize.
  fseek(inFile, 0L, SEEK_END);
  size_t fileSize = ftell(inFile);
  fseek(inFile, 0L, SEEK_SET);
  std::cout << "fileSize = " << fileSize << " bytes (about " << ((double)fileSize)/1000000000 << " GB)." << std::endl;
  // Read file contents into buffer
  const int BUFSIZE = fileSize + 10000;
  std::vector<char> buf(BUFSIZE);
  memset(&buf[0], 0x00, BUFSIZE);
  if(fread(&buf[0], 1, BUFSIZE, inFile) <= 0)
    throw std::runtime_error("Error reading inFile.");
  if(buf[BUFSIZE-1] != '\0')
    throw std::runtime_error("Error: zero not found at end of buffer. This should not happen.");
  fclose(inFile);
  // Count number of newline characters
  int newlineCount = 0;
  for(int i = 0; i < fileSize; i++) {
    if(buf[i] == '\n')
      newlineCount++;
  }
  size_t seqSize = fileSize - newlineCount;
  resultSeq.resize(seqSize);
  int k = 0;
  for(int i = 0; i < fileSize; i++) {
    if(buf[i] != '\n')
      resultSeq[k++] = buf[i];
  }
}

static std::string get_seq_string(const std::vector<char> & seq) {
  int n = seq.size();
  char s[n+1];
  for(int i = 0; i < n; i++)
    s[i] = seq[i];
  s[n] = '\0';
  return (std::string)s;
}

static int max4(int a1, int a2, int a3, int a4) {
  if(a1 > a2 && a1 > a3 && a1 > a4)
    return a1;
  if(a2 > a3 && a2 > a4)
    return a2;
  if(a3 > a4)
    return a3;
  return a4;
}

int main(int argc, const char* argv[]) {
  if(argc != 3) {
    std::cout << "Please give two arguments: input file names for the two sequences that are to be aligned." << std::endl;
    return -1;
  }
  const char* fileNameA = argv[1];
  const char* fileNameB = argv[2];
  std::vector<char> seqA;
  std::vector<char> seqB;
  read_sequence_from_file(fileNameA, seqA);
  read_sequence_from_file(fileNameB, seqB);
  assert(seqA.size() == seqB.size());
  int N = seqA.size();
  
  printf("Sequence A : %s\n", get_seq_string(seqA).c_str());
  printf("Sequence B : %s\n", get_seq_string(seqB).c_str());

  const int similarity_score_for_match = 2;
  const int similarity_score_for_mismatch = -1;
  const int gap_penalty_d = 2;
  
  std::vector<int> H(N*N);
  for(int startRow = 0; startRow < 2*N-1; startRow++) {
    for(int col = 0; col <= startRow; col++) {
      int row = startRow - col;
      assert(row >= 0);
      if(row >= N)
	continue;
      if(col >= N)
	continue;
      int H_topleft = 0;
      if(row > 0 && col > 0)
	H_topleft = H[(row-1)*N+(col-1)];
      int H_top = 0;
      if(row > 0)
	H_top = H[(row-1)*N+(col-0)];
      int H_left = 0;
      if(col > 0)
	H_left = H[(row-0)*N+(col-1)];
      int S_similarity_score = similarity_score_for_mismatch;
      if(seqA[row] == seqB[col])
	S_similarity_score = similarity_score_for_match;
      // Take max of 4 alternatives
      int alt1 = H_topleft + S_similarity_score;
      int alt2 = H_left - gap_penalty_d;
      int alt3 = H_top - gap_penalty_d;
      int alt4 = 0;
      H[row*N+col] = max4(alt1, alt2, alt3, alt4);
    }
  }

  printf("H:\n");
  for(int i = 0; i < N; i++) {
    for(int j = 0; j < N; j++)
      printf("%3d ", H[j*N+i]);
    printf("\n");
  }
  
  std::cout << "Done." << std::endl;

  return 0;
}

