/* Sample code using the Chunks and Tasks programming model.
 * See <http://www.chunks-and-tasks.org>.
 */
#include <iostream>
#include <cstdio>
#include <cstring>
#include <cassert>
#include <vector>
#include "SmithWaterman.h"

static void read_sequence_from_file(const char* fileName, std::vector<char> & resultSeq) {
  FILE* inFile = fopen(fileName, "rb");
  if(!inFile)
    throw std::runtime_error("Error opening inFile.");
  // Get filesize.
  fseek(inFile, 0L, SEEK_END);
  size_t fileSize = ftell(inFile);
  fseek(inFile, 0L, SEEK_SET);
  std::cout << "fileSize = " << fileSize << " bytes (about " << ((double)fileSize)/1000000000 << " GB)." << std::endl;
  // Read file contents into buffer
  const int BUFSIZE = fileSize + 10000;
  std::vector<char> buf(BUFSIZE);
  memset(&buf[0], 0x00, BUFSIZE);
  if(fread(&buf[0], 1, BUFSIZE, inFile) <= 0)
    throw std::runtime_error("Error reading inFile.");
  if(buf[BUFSIZE-1] != '\0')
    throw std::runtime_error("Error: zero not found at end of buffer. This should not happen.");
  fclose(inFile);
  // Count number of newline characters
  int newlineCount = 0;
  for(int i = 0; i < fileSize; i++) {
    if(buf[i] == '\n')
      newlineCount++;
  }
  size_t seqSize = fileSize - newlineCount;
  resultSeq.resize(seqSize);
  int k = 0;
  for(int i = 0; i < fileSize; i++) {
    if(buf[i] != '\n')
      resultSeq[k++] = buf[i];
  }
}

static std::string get_seq_string(const std::vector<char> & seq) {
  int n = seq.size();
  char s[n+1];
  for(int i = 0; i < n; i++)
    s[i] = seq[i];
  s[n] = '\0';
  return (std::string)s;
}

static int max4(int a1, int a2, int a3, int a4) {
  if(a1 > a2 && a1 > a3 && a1 > a4)
    return a1;
  if(a2 > a3 && a2 > a4)
    return a2;
  if(a3 > a4)
    return a3;
  return a4;
}

int main(int argc, const char* argv[]) {
  if(argc != 3) {
    std::cout << "Please give two arguments: input file names for the two sequences that are to be aligned." << std::endl;
    return -1;
  }
  const char* fileNameA = argv[1];
  const char* fileNameB = argv[2];
  std::vector<char> seqA;
  std::vector<char> seqB;
  read_sequence_from_file(fileNameA, seqA);
  read_sequence_from_file(fileNameB, seqB);
  assert(seqA.size() == seqB.size());
  int N = seqA.size();
  
  printf("Sequence A : %s\n", get_seq_string(seqA).c_str());
  printf("Sequence B : %s\n", get_seq_string(seqB).c_str());

  const int similarity_score_for_match = 2;
  const int similarity_score_for_mismatch = -1;
  const int gap_penalty_d = 2;
  
  std::vector<int> H(N*N);
  int maxValue_serial = 0;
  for(int startRow = 0; startRow < 2*N-1; startRow++) {
    for(int col = 0; col <= startRow; col++) {
      int row = startRow - col;
      assert(row >= 0);
      if(row >= N)
	continue;
      if(col >= N)
	continue;
      int H_topleft = 0;
      if(row > 0 && col > 0)
	H_topleft = H[(row-1)*N+(col-1)];
      int H_top = 0;
      if(row > 0)
	H_top = H[(row-1)*N+(col-0)];
      int H_left = 0;
      if(col > 0)
	H_left = H[(row-0)*N+(col-1)];
      int S_similarity_score = similarity_score_for_mismatch;
      if(seqA[row] == seqB[col])
	S_similarity_score = similarity_score_for_match;
      // Take max of 4 alternatives
      int alt1 = H_topleft + S_similarity_score;
      int alt2 = H_left - gap_penalty_d;
      int alt3 = H_top - gap_penalty_d;
      int alt4 = 0;
      int Hvalue = max4(alt1, alt2, alt3, alt4);
      H[row*N+col] = Hvalue;
      if(Hvalue > maxValue_serial)
	maxValue_serial = Hvalue;
    }
  }
  printf("H (computed serially):\n");
  for(int row = 0; row < N; row++) {
    for(int col = 0; col < N; col++)
      printf("%3d ", H[row*N+col]);
    printf("\n");
  }  
  std::cout << "Serial SW done." << std::endl;
  printf("maxValue_serial = %d\n", maxValue_serial);

  // CHT part starts here

  cht::start();

  std::vector<int> v_left_border(N);
  std::vector<int> v_top_border(N);
  for(int i = 0; i < N; i++) {
    v_left_border[i] = 0;
    v_top_border[i] = 0;
  }

  cht::ChunkID cid_seq_A = cht::registerChunk<chttl::ChunkVector<char> >(new chttl::ChunkVector<char>(seqA));
  cht::ChunkID cid_seq_B = cht::registerChunk<chttl::ChunkVector<char> >(new chttl::ChunkVector<char>(seqB));
  cht::ChunkID cid_left_border = cht::registerChunk<chttl::ChunkVector<int> >(new chttl::ChunkVector<int>(v_left_border));
  cht::ChunkID cid_top_border = cht::registerChunk<chttl::ChunkVector<int> >(new chttl::ChunkVector<int>(v_top_border));
  cht::ChunkID cid_top_left_value = cht::registerChunk<chttl::Basic<int> >(new chttl::Basic<int>(0));

#if 1
  cht::ChunkID cid_result  = cht::executeMotherTask<SW>(cid_seq_A,
							cid_seq_B,
							cid_left_border,
							cid_top_border,
							cid_top_left_value);
  cht::shared_ptr<SWResult const> ptr_result;
  cht::getChunk(cid_result, ptr_result);
  std::cout << "CHT SW done." << std::endl;
  std::cout << "CHT computed result maxValue : " << ptr_result->maxValue << std::endl;
  assert(ptr_result->maxValue == maxValue_serial);
#endif

  cht::deleteChunk(cid_seq_A);
  cht::deleteChunk(cid_seq_B);
  cht::deleteChunk(cid_left_border);
  cht::deleteChunk(cid_top_border);
  cht::deleteChunk(cid_top_left_value);

  cht::deleteChunk(cid_result);

  cht::stop();

  printf("Smith-Waterman test finished OK!\n");
  return 0;
}

